#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "s3vn.h"

int main(int argc, char **argv)
{

  int rc;
  char *repo_name, *current_rev;
  //int current_local, current_arg, txn_local, current_zoo;
  int current_local, current_arg, current_zoo;

  // Get the local repository path from the command line arguments
  if (argc < 3)
  {
    printf("Usage: post-commit-hook <repos_path> <rev>\n");
    return 1;
  }
  repo_name = argv[1];
  current_rev = argv[2];

  printf("Repository name is %s\n", repo_name);

  // Initialize s3vn configuration
  init_s3vn( repo_name,
             get_env_or_die("SVNZOO_BUCKET"),
             get_env_or_die("SVNZOO_PREFIX"),
             get_env_or_die("SVNZOO_ZKADDR"),
             get_env_or_die("SVNZOO_ZKPREFIX")
           );

  // Connect to ZooKeeper
  rc = restore_session();
  if (rc == -1)
  {
    fprintf(stderr, "Could not reconnect to ZooKeeper\n");
    exit(1);
  }

  // Get local version info
  current_local = get_current();
  //txn_local = get_txn();
  //if (current_local == -1 || txn_local == -1)
  if (current_local == -1)
  {
    //fprintf(stderr, "Error in (current_local, txn_local): (%d, %d)\n", current_local, txn_local);
    fprintf(stderr, "Error in (current_local): (%d)\n", current_local);
    release_lock(); // ignore any errors (actually may not be necessary at all, if lock is ephemeral)
    end_session();
    exit(1);
  }

  // Some sanity checks, for good measure...
  if (sscanf(current_rev, "%d", &current_arg) != 1)
  {
    fprintf(stderr, "Sanity check failed. Local revision is %d, but argument is %d\n", current_local, current_arg);
    // TODO: reinstate the following lines when not testing
    //release_lock(); // ignore any errors (actually may not be necessary at all, if lock is ephemeral)
    //end_session();
    //exit(1);
  }

  // Make sure we still hold the lock. This is more if a sanity check to mitigate damage (or possibly
  // recover) in case something goes terribly wrong, since ZooKeeper should NEVER lose the lock on us.
  rc = check_lock();
  if (rc == -1)
  {
    fprintf(stderr, "Could not determine whether the lock is still held\n");
    release_lock(); // ignore any errors (actually may not be necessary at all, if lock is ephemeral)
    end_session();
    exit(1);
  }
  else if (rc == 0)
  {
    // !Ay caramba! We've lost the lock...
    fprintf(stderr, "Trying to re-acquire the lock, which was somehow lost\n");
    rc = acquire_lock();
    if (rc == -1)
    {
      fprintf(stderr, "Could not re-acquire lock\n");
      end_session();
      exit(1);
    }
    current_zoo = fetch_current();
    if (current_zoo == -1)
    {
      fprintf(stderr, "Could not determine latest ZooKeeper revision number\n");
      release_lock(); // ignore any errors (actually may not be necessary at all, if lock is ephemeral)
      end_session();
      exit(1);
    }
    if (current_zoo != (current_local-1))
    {
      fprintf(stderr, "Aargh! Somebody commited a newer revision while during the time we lost the lock!\n");
      release_lock(); // ignore any errors (actually may not be necessary at all, if lock is ephemeral)
      end_session();
      exit(1);
    }
  }

  // Push the new revision into S3
  rc = store_rev(current_local);
  if (rc == -1)
  {
    fprintf(stderr, "Could not push revision %d into S3\n", current_local);
    release_lock(); // ignore any errors (actually may not be necessary at all, if lock is ephemeral)
    end_session();
    exit(1);
  }

  // Update current and txn in ZooKeeper
  rc = store_current(current_local);
  if (rc == -1)
  {
    fprintf(stderr, "Could not push current revision number %d into ZooKeeper\n", current_local);
    release_lock(); // ignore any errors (actually may not be necessary at all, if lock is ephemeral)
    end_session();
    exit(1);
  }
  //rc = store_txn(txn_local);
  //if (rc == -1)
  //{
  //  fprintf(stderr, "Could not push transaction number %d into ZooKeeper\n", txn_local);
  //  release_lock(); // ignore any errors (actually may not be necessary at all, if lock is ephemeral)
  //  end_session();
  //  exit(1);
  //}

  // Delete the information about the current session from the local repository
  rc = delete_session();
  if (rc == -1)
  {
    fprintf(stderr, "Could not delete session info\n");
    release_lock(); // ignore any errors (actually may not be necessary at all, if lock is ephemeral)
    end_session();
    exit(1);
  }

  // Release the repository write lock in ZooKeeper
  rc = release_lock();
  if (rc == -1)
  {
    fprintf(stderr, "Could not release repository lock in ZooKeeper\n");
    end_session();
    exit(1);
  }

  // We are done with the ZooKeeper session!
  end_session();

  return 0;
}
