#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <pthread.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include <c-client-src/zookeeper.h>

#include "s3vn.h"

#define SVNADMIN_CMD "svnadmin create --fs-type=fsfs --pre-1.6-compatible %s"
#define ZOOKEEPER_TIMEOUT 10000
#define SHARD_COUNT 1000
#define S3_RETRY_TIME 2
#define LOG_LEVEL_FOR_ZOO ZOO_LOG_LEVEL_WARN
#define PATH_BUF_LEN 128
#define LONG_PATH_BUF_LEN 1024
#define STRING_BUF_LEN 128  // for revision number string
#define LOCK_FORMAT_STR "%010u"
#define SHELL_CMD_BUF_LEN 512

// Module-level global variables
static struct s3vn_config configuration;
static zhandle_t* zh;
static char *progname = "new";

// Initialize the s3vn module
void init_s3vn(char* local_repo, char* s3_bucket, char* s3_prefix, char* zk_addr, char* zk_prefix)
{
  configuration.local_repo = local_repo;
  configuration.s3_bucket = s3_bucket;
  configuration.s3_prefix = s3_prefix;
  configuration.zk_addr = zk_addr;
  configuration.zk_prefix = zk_prefix;
}

// Set program name (for preventing temporary file conflicts)
void set_progname(char *pn)
{
  progname = pn;
}

// Create an empty repository on the local disk
int create_local_repo(void)
{
  int rc;
  char cmd[SHELL_CMD_BUF_LEN];
  cmd[SHELL_CMD_BUF_LEN-1] = '\0';

  snprintf(cmd, SHELL_CMD_BUF_LEN-1, SVNADMIN_CMD, configuration.local_repo);
  rc = system(cmd);
  if (rc == -1)
  {
    perror("could not execute svnadmin command");
    return -1;
  }
  else if (rc != 0)
  {
    fprintf(stderr, "svnadmin failed and returned value %d\n", rc);
    return -1;
  }

  // Install the start-commit hook script
  snprintf(cmd, SHELL_CMD_BUF_LEN-1,
           "s3x get %s s3vn-start-commit.sh %s/hooks/start-commit && chmod 755 %s/hooks/start-commit",
           get_env_or_die("BUCKET_NAME"), configuration.local_repo, configuration.local_repo);
  rc = system(cmd);
  if (rc == -1)
  {
    perror("could not get s3vn-start-commit.sh from S3");
    return -1;
  }
  else if (rc != 0)
  {
    fprintf(stderr, "could not get s3vn-start-commit.sh from S3\n");
    return -1;
  }

  // Install the post-commit hook script
  snprintf(cmd, SHELL_CMD_BUF_LEN-1,
           "s3x get %s s3vn-post-commit.sh %s/hooks/post-commit && chmod 755 %s/hooks/post-commit",
           get_env_or_die("BUCKET_NAME"), configuration.local_repo, configuration.local_repo);
  rc = system(cmd);
  if (rc == -1)
  {
    perror("could not get s3vn-post-commit.sh from S3");
    return -1;
  }
  else if (rc != 0)
  {
    fprintf(stderr, "could not get s3vn-post-commit.sh from S3\n");
    return -1;
  }

  return 0;
}

static char * _uuid_filename()
{
  static char uuid_file[LONG_PATH_BUF_LEN];
  uuid_file[LONG_PATH_BUF_LEN-1] = '\0';
  snprintf(uuid_file, LONG_PATH_BUF_LEN-1, "%s/db/uuid", configuration.local_repo);
  return uuid_file;
}

static char * _uuid_key()
{
  static char uuid_key[PATH_BUF_LEN];
  uuid_key[PATH_BUF_LEN-1] = '\0';
  snprintf(uuid_key, PATH_BUF_LEN-1, "%suuid", configuration.s3_prefix);
  return uuid_key;
}

// Fetch a repository from S3/ZooKeeper
int fetch_repo(void)
{
  //int current_rev, current_txn, rc, uuid_len;
  int current_rev, rc;
  char buf[SHELL_CMD_BUF_LEN];

  current_rev = fetch_current();
  //current_txn = fetch_txn();
  //if (current_rev == -1 || current_txn == -1)
  if (current_rev == -1)
  {
    fprintf(stderr, "Cannot fetch repository\n");
    return -1;
  }
  rc = set_current(current_rev);
  if (rc == -1)
  {
    fprintf(stderr, "Cannot set current version\n");
    return -1;
  }
  //rc = set_txn(current_txn);
  //if (rc == -1)
  //{
  //  fprintf(stderr, "Cannot set transaction number\n");
  //  return -1;
  //}

  // Get the UUID from S3 and put it into the repository
  snprintf(buf, sizeof(buf)-1, "s3x get %s %s %s", configuration.s3_bucket, _uuid_key(), _uuid_filename());
  buf[sizeof(buf)-1] = '\0';
  rc = system(buf);
  if (rc != 0)
  {
    fprintf(stderr, "Could not get uuid file from S3\n");
    return -1;
  }

  // Get the actual revisions
  for (int i=0; i<=current_rev; i++)
  {
    if ( fetch_rev(i) == -1 )
      return -1;
  }

  return 0;
}

static char * _lock_node()
{
  static char lock_node[PATH_BUF_LEN];
  lock_node[PATH_BUF_LEN-1] = '\0';
  snprintf(lock_node, PATH_BUF_LEN-1, "%s/lock", configuration.zk_prefix);
  return lock_node;
}

static char * _lock_path()
{
  static char lock_path[PATH_BUF_LEN];
  lock_path[PATH_BUF_LEN-1] = '\0';
  snprintf(lock_path, PATH_BUF_LEN-1, "%s/lock-", _lock_node());
  return lock_path;
}

// Store a local repository info S3/ZooKeeper
int store_repo(void)
{
  //int current_rev, current_txn, rc;
  int current_rev, rc;
  FILE* fd;
  int fdno;
  char uuid_buf[100];
  char buf[SHELL_CMD_BUF_LEN];
  int uuid_buf_len = sizeof(uuid_buf);
  int uuid_buf_pos = 0;
  int created = 0;

  // Get the current revision and transaction numbers
  current_rev = get_current();
  //current_txn = get_txn();
  //if (current_rev < 0 || current_txn < 0)
  if (current_rev < 0)
  {
    fprintf(stderr, "Error in getting either the revision or transaction number\n");
    return -1;
  }

  // Load the UUID from disk
  fd = fopen(_uuid_filename(), "r");
  if (!fd)
  {
    perror("Cannot open UUID file for reading");
    return -1;
  }
  fdno = fileno(fd);
  while (1)
  {
    rc = read(fdno, uuid_buf+uuid_buf_pos, uuid_buf_len-uuid_buf_pos);
    if (rc < 0)
    {
      perror("Cannot read from UUID file");
      return -1;
    }
    else if (rc == 0)
      break;
    else
      uuid_buf_pos += rc;
  }
  fclose(fd);

  // Store the UUID into S3
  snprintf(buf, sizeof(buf)-1, "s3x put %s %s %s", configuration.s3_bucket, _uuid_key(), _uuid_filename());
  buf[sizeof(buf)-1] = '\0';
  rc = system(buf);
  if (rc != 0)
  {
    fprintf(stderr, "Could not put uuid file into S3\n");
    return -1;
  }

  // Store each revision into S3
  for (int i=0; i<=current_rev; i++)
  {
    if ( store_rev(i) == -1 )
      return -1;
  }

  // Make sure the ZooKeeper prefix node exists and store the UUID into it
  rc = zoo_exists(zh, configuration.zk_prefix, 0, 0);
  if (rc == ZNONODE)
  {
    // Create the prefix node
    rc = zoo_create(zh, configuration.zk_prefix, uuid_buf, uuid_buf_pos, &ZOO_OPEN_ACL_UNSAFE, 0, 0, 0);
    if (rc != ZOK)
    {
      fprintf(stderr, "Could not create ZooKeeper prefix node: %s\n", zerror(rc));
      return -1;
    }
    created = 1;
  }
  else if (rc != ZOK)
  {
    fprintf(stderr, "Could not determine if ZooKeeper prefix node exists\n");
    return -1;
  }
  if (!created)
  {
    // Since the prefix node already existed, we need to push the current UUID into it
    rc = zoo_set(zh, configuration.zk_prefix, uuid_buf, uuid_buf_pos, -1);
    if (rc != ZOK)
    {
      fprintf(stderr, "Could not push UUID into ZooKeeper prefix node: %s\n", zerror(rc));
      return -1;
    }
  }

  // Create a ZooKeeper lock node
  rc = zoo_create(zh, _lock_node(), "", 0, &ZOO_OPEN_ACL_UNSAFE, 0, 0, 0);
  if (rc != ZOK && rc != ZNODEEXISTS)
  {
      fprintf(stderr, "Could not create ZooKeeper lock node: %s\n", zerror(rc));
      return -1;
  }

  // Store the current revision and txn into ZooKeeper
  rc = store_current(current_rev);
  if (rc == -1) return -1;
  //rc = store_txn(current_txn);
  //if (rc == -1) return -1;

  return 0;
}

static char * _rev_filename(int revnum)
{
  int shard;
  static char path[LONG_PATH_BUF_LEN];
  path[LONG_PATH_BUF_LEN-1] = '\0';
  shard = revnum / SHARD_COUNT;
  snprintf(path, LONG_PATH_BUF_LEN-1, "%s/db/revs/%d/%d", configuration.local_repo, shard, revnum);
  return path;
}

static char * _rev_filename_tmp(int revnum)
{
  int shard;
  static char path[LONG_PATH_BUF_LEN];
  path[LONG_PATH_BUF_LEN-1] = '\0';
  shard = revnum / SHARD_COUNT;
  snprintf(path, LONG_PATH_BUF_LEN-1, "%s/db/revs/%d/%d.%s", configuration.local_repo, shard, revnum, progname);
  return path;
}

static char * _revprop_filename(int revnum)
{
  int shard;
  static char path[LONG_PATH_BUF_LEN];
  path[LONG_PATH_BUF_LEN-1] = '\0';
  shard = revnum / SHARD_COUNT;
  snprintf(path, LONG_PATH_BUF_LEN-1, "%s/db/revprops/%d/%d", configuration.local_repo, shard, revnum);
  return path;
}

static char * _revprop_filename_tmp(int revnum)
{
  int shard;
  static char path[LONG_PATH_BUF_LEN];
  path[LONG_PATH_BUF_LEN-1] = '\0';
  shard = revnum / SHARD_COUNT;
  snprintf(path, LONG_PATH_BUF_LEN-1, "%s/db/revprops/%d/%d.%s", configuration.local_repo, shard, revnum, progname);
  return path;
}

static char * _rev_shard_dir(int revnum)
{
  static char path[LONG_PATH_BUF_LEN];
  path[LONG_PATH_BUF_LEN-1] = '\0';
  snprintf(path, LONG_PATH_BUF_LEN-1, "%s/db/revs/%d", configuration.local_repo, revnum / SHARD_COUNT);
  return path;
}

static char * _revprop_shard_dir(int revnum)
{
  static char path[LONG_PATH_BUF_LEN];
  path[LONG_PATH_BUF_LEN-1] = '\0';
  snprintf(path, LONG_PATH_BUF_LEN-1, "%s/db/revprops/%d", configuration.local_repo, revnum / SHARD_COUNT);
  return path;
}

static char * _rev_key(int revnum)
{
  static char s3_key[PATH_BUF_LEN];
  s3_key[PATH_BUF_LEN-1] = '\0';
  snprintf(s3_key, PATH_BUF_LEN-1, "%srev-%d", configuration.s3_prefix, revnum);
  return s3_key;
}

static char * _revprop_key(int revnum)
{
  static char s3_key[PATH_BUF_LEN];
  s3_key[PATH_BUF_LEN-1] = '\0';
  snprintf(s3_key, PATH_BUF_LEN-1, "%srevprop-%d", configuration.s3_prefix, revnum);
  return s3_key;
}

// Fetch the specified revision from S3 to the local repository
int fetch_rev(int revnum)
{
  char *s3_key;
  int rc;
  char buf[SHELL_CMD_BUF_LEN];

  // Determine the S3 key for the revision data and the local filename store it to
  s3_key = _rev_key(revnum);
  snprintf(buf, sizeof(buf)-1, "s3x get %s %s %s", configuration.s3_bucket, s3_key, _rev_filename_tmp(revnum));
  buf[sizeof(buf)-1] = '\0';

  // Create the shard directory for the revision data in case it doesn't exist
  rc = mkdir(_rev_shard_dir(revnum), 0755);
  if (rc == -1 && errno != EEXIST)
  {
    fprintf(stderr, "Could not create shard directory for revision data %d\n", revnum);
    return -1;
  }

  // Fetch the revision data from S3
  while (1)
  {
    rc = system(buf);
    if (rc != 0)  // FIXME: should handle an S3 read error differently from a local write error!
    {
      // The fetch failed, but this function should only ever have been called if ZooKeeper
      // told that the revision does in fact exist. So probably Amazon just hasn't propagated
      // it to all of the S3 hosts yet. There's nothing we can do but wait.
      sleep(S3_RETRY_TIME); // Could be fancier here, using exponentially increasing times
                            // and eventually timing out.
      continue;
    }
    break;
  }

  // Create the shard directory for the revision property in case it doesn't exist
  rc = mkdir(_revprop_shard_dir(revnum), 0755);
  if (rc == -1 && errno != EEXIST)
  {
    fprintf(stderr, "Could not create shard directory for revision property %d\n", revnum);
    return -1;
  }

  // Determine the S3 key for the revision property and the local filename to store it to
  s3_key = _revprop_key(revnum);
  snprintf(buf, sizeof(buf)-1, "s3x get %s %s %s", configuration.s3_bucket, s3_key, _revprop_filename_tmp(revnum));
  buf[sizeof(buf)-1] = '\0';

  // Fetch the revision property from S3
  while (1)
  {
    rc = system(buf);
    if (rc != 0)
    {
      // Keep trying, see comment above
      sleep(S3_RETRY_TIME);
      continue;
    }
    break;
  }

  // Rename the files into place
  rc = rename(_rev_filename_tmp(revnum), _rev_filename(revnum));
  if (rc != 0)
  {
    perror("Could not rename revision data file into place");
    return -1;
  }
  rc = rename(_revprop_filename_tmp(revnum), _revprop_filename(revnum));
  if (rc != 0)
  {
    perror("Could not rename revision property file into place");
    return -1;
  }

  return 0;
}

// Push the specified revision from the local repository into S3
int store_rev(int revnum)
{
  char *path, *s3_key;
  int rc;
  char buf[SHELL_CMD_BUF_LEN];

  // Determine the local path of the revision data
  path = _rev_filename(revnum);

  // Determine the S3 path for the revision data
  s3_key = _rev_key(revnum);

  // Store the revision data into S3
  snprintf(buf, sizeof(buf)-1, "s3x put %s %s %s", configuration.s3_bucket, s3_key, path);
  buf[sizeof(buf)-1] = '\0';
  while (1) // retry until successful
  {
    rc = system(buf);
    if (rc == 0) break;
    fprintf(stderr, "could not put revision %d into S3\n", revnum);
    sleep(1);
  }

  // Determine the local path of the revision properties
  path = _revprop_filename(revnum);

  // Determine the S3 path for the revision properties
  s3_key = _revprop_key(revnum);

  // Store the revision properties into S3
  snprintf(buf, sizeof(buf)-1, "s3x put %s %s %s", configuration.s3_bucket, s3_key, path);
  buf[sizeof(buf)-1] = '\0';
  while (1) // retry until successful
  {
    rc = system(buf);
    if (rc == 0) break;
    fprintf(stderr, "could not put revision property %d into S3\n", revnum);
    sleep(1);
  }

  return 0;
}

static char * _get_current_file_name()
{
  char *current_file;
  int rc;

  rc = asprintf(&current_file, "%s/db/current", configuration.local_repo);
  if (rc == -1)
  {
    fprintf(stderr, "could not allocate memory in asprintf\n");
    exit(1);
  }

  return current_file;
}

static char * _get_current_file_name_tmp()
{
  char *current_file;
  int rc;

  rc = asprintf(&current_file, "%s/db/current.%s", configuration.local_repo, progname);
  if (rc == -1)
  {
    fprintf(stderr, "could not allocate memory in asprintf\n");
    exit(1);
  }

  return current_file;
}

static int _set_string(char *file, char *tmpfile, char *str)
{
  FILE *fd;
  int rc;

  fd = fopen(tmpfile, "w");
  if (!fd)
  {
    perror("Could not write string to local repository");
    return -1;
  }

  rc = fprintf(fd, "%s\n", str);
  if (rc < 0)
  {
    fprintf(stderr, "Could not write to string file\n");
    fclose(fd);
    return -1;
  }

  fclose(fd);

  rc = rename(tmpfile, file);
  if (rc == -1)
  {
    perror("Could not put new file in place by a rename");
    return -1;
  }

  return 0;
}

static int _set_integer(char *file, char *tmpfile, int the_int)
{
  FILE *fd;
  int rc;

  fd = fopen(tmpfile, "w");
  if (!fd)
  {
    perror("Could not write current revision number to local repository");
    return -1;
  }

  rc = fprintf(fd, "%d\n", the_int);
  if (rc < 0)
  {
    fprintf(stderr, "Could not write to revision number file\n");
    fclose(fd);
    return -1;
  }

  fclose(fd);

  rc = rename(tmpfile, file);
  if (rc == -1)
  {
    perror("Could not put new file in place by a rename");
    return -1;
  }

  return 0;
}

// Set the current revision number in the local repository
int set_current(int revnum)
{
  char *current_file, *current_file_tmp;
  int rc;

  current_file = _get_current_file_name();
  current_file_tmp = _get_current_file_name_tmp();

  rc = _set_integer(current_file, current_file_tmp, revnum);

  free(current_file);
  free(current_file_tmp);

  return rc;
}

static char * _get_string(char *file)
{
  FILE *fd;
  int rc;
  static char rev_num[STRING_BUF_LEN];

  fd = fopen(file, "r");
  if (!fd)
  {
    perror("Could not read string from local repository");
    return 0;
  }

  rc = fscanf(fd, "%s", rev_num); // FIXME: should make sure there isn't a buffer overrun here
  if (rc != 1)
  {
    fprintf(stderr, "could not parse string from file\n");
    fclose(fd);
    return 0;
  }

  fclose(fd);

  return rev_num;
}

static int _get_integer(char *file)
{
  FILE *fd;
  int rc, rev_num;

  fd = fopen(file, "r");
  if (!fd)
  {
    perror("Could not read current revision number from local repository");
    return -1;
  }

  rc = fscanf(fd, "%d", &rev_num);
  if (rc != 1)
  {
    fprintf(stderr, "could not parse revision number from current file\n");
    fclose(fd);
    return -1;
  }

  fclose(fd);

  return rev_num;
}

// Get the current revision number from the local repository
int get_current(void)
{
  char *current_file;
  int rc;

  current_file = _get_current_file_name();

  rc = _get_integer(current_file);

  free(current_file);

  return rc;
}

static char * _get_txn_file_name(int tmp)
{
  static char txn_file[LONG_PATH_BUF_LEN];
  txn_file[LONG_PATH_BUF_LEN-1] = '\0';

  snprintf(txn_file, LONG_PATH_BUF_LEN-1, "%s/db/txn-current%s", configuration.local_repo, tmp ? ".new" : "");

  return txn_file;
}

// Set the transaction number in the local repository
int set_txn(char *txnnum)
{
  char *txn_file, *txn_file_tmp;
  int rc;

  txn_file = _get_txn_file_name(0);
  txn_file_tmp = _get_txn_file_name(1);

  

  rc = _set_string(txn_file, txn_file_tmp, txnnum);

  free(txn_file);
  free(txn_file_tmp);

  return rc;
}

// Get the transaction number from the local repository
int get_txn(void)
{
  char *txn_file;
  int rc;

  txn_file = _get_txn_file_name(0);

  rc = _get_integer(txn_file);

  free(txn_file);

  return rc;
}

static int _store_int(const char *path, int v)
{
  int rc, ilen;
  char ibuf[32];
  int ibuf_len = sizeof(ibuf);

  ilen = snprintf(ibuf, ibuf_len, "%d", v);
  if (ilen < 1) return -1;
  rc = zoo_set(zh, path, ibuf, ilen, -1);
  if (rc == ZNONODE)
  {
    rc = zoo_create(zh, path, ibuf, ilen, &ZOO_OPEN_ACL_UNSAFE, 0, 0, 0);
    // Race condition here if somebody else creates the node, but this should never happen in the s3vn application
    if (rc != ZOK)
    {
      fprintf(stderr, "Could not create node for storing integer in ZooKeeper: %s\n", zerror(rc));
      return -1;
    }
  }
  else if (rc != ZOK)
  {
    fprintf(stderr, "Could not store integer into ZooKeeper: %s\n", zerror(rc));
    return -1;
  }
  return 0;
}

static int _fetch_int(const char *path)
{
  int rc, the_int;
  char buf[100];
  int sz = sizeof(buf) - 1;

  rc = zoo_get(zh, path, 0, buf, &sz, 0);
  if (rc != ZOK)
  {
    fprintf(stderr, "Could not get integer from ZooKeeper: %s\n", zerror(rc));
    return -1;
  }

  buf[sz] = '\0';
  rc = sscanf(buf, "%d", &the_int);
  if (rc != 1)
  {
    fprintf(stderr, "Could not parse integer from ZooKeeper\n");
    return -1;
  }

  return the_int;
}

static char * _zk_current_path()
{
  static char path[PATH_BUF_LEN];
  path[PATH_BUF_LEN-1] = '\0';
  snprintf(path, PATH_BUF_LEN-1, "%s/current", configuration.zk_prefix);
  return path;
}

// Push the current revision number into ZooKeeper
int store_current(int revnum)
{
  char *path;
  int rc;

  path = _zk_current_path();

  rc = _store_int(path, revnum);
  if (rc == -1)
  {
    fprintf(stderr, "Could not store current revision number into ZooKeeper\n");
    return -1;
  }

  return 0;
}

// Get the number of the latest revision that ZooKeeper knows about
int fetch_current(void)
{
  char *path;
  int rc;

  path = _zk_current_path();

  rc = _fetch_int(path);
  if (rc == -1)
  {
    fprintf(stderr, "Could not fetch current revision number from ZooKeeper\n");
    return -1;
  }

  return rc;
}

static pthread_mutex_t _current_watcher_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t _current_watcher_cond = PTHREAD_COND_INITIALIZER;

static void _current_watcher(zhandle_t *zzh, int type, int state, const char *path, void *watcherCtx)
{
  pthread_mutex_lock(&_current_watcher_mutex); // grab the mutex
  pthread_cond_signal(&_current_watcher_cond); // wake the thread that is waiting
  pthread_mutex_unlock(&_current_watcher_mutex); // unlock
}

// Wait until a revision greater than current_rev is available, then return it
int watch_current(int current_rev)
{
  int rc, latest_rev, sz;
  char buf[512];
  struct Stat stat;

  pthread_mutex_lock(&_current_watcher_mutex); // must grab mutex before setting a watch
  // WARNING: must be absolutely sure to unlock on EVERY possible exit path from the loop

  while (1)
  {

    // Retrieve the current value and set a watch
    sz = sizeof(buf) - 1;
    rc = zoo_wget(zh, _zk_current_path(), _current_watcher, 0, buf, &sz, &stat);
    if (rc != ZOK)
    {
      fprintf(stderr, "Could not get current rev from ZooKeeper in watcher function: %s\n", zerror(rc));
      pthread_mutex_unlock(&_current_watcher_mutex); // be sure to unlock
      return -1;
    }
    buf[sz] = '\0';
    rc = sscanf(buf, "%d", &latest_rev);
    if (rc != 1)
    {
      fprintf(stderr, "Could not parse integer from ZooKeeper\n");
      pthread_mutex_unlock(&_current_watcher_mutex); // be sure to unlock
      return -1;
    }

    // Return the latest rev number if it is greater than the current
    if (latest_rev > current_rev)
    {
      pthread_mutex_unlock(&_current_watcher_mutex); // unlock
      return latest_rev;
    }

    // Otherwise wait for a notification
    pthread_cond_wait(&_current_watcher_cond, &_current_watcher_mutex);

  }

  // Should never get here!
  return -1;
}

static char * _zk_txn_path()
{
  int rc;
  char *path;
  rc = asprintf(&path, "%s/txn", configuration.zk_prefix);
  if (rc == -1)
  {
    fprintf(stderr, "could not allocate memory in asprintf\n");
    exit(1);
  }
  return path;
}

// Push the current transaction number into ZooKeeper
int store_txn(int txnnum)
{
  char *path;
  int rc;

  path = _zk_txn_path();

  rc = _store_int(path, txnnum);
  free(path);
  if (rc == -1)
  {
    fprintf(stderr, "Could not store transaction number into ZooKeeper\n");
    return -1;
  }

  return 0;
}

// Get the latest transaction number that ZooKeeper knows about
int fetch_txn(void)
{
  char *path;
  int rc;

  path = _zk_txn_path();

  rc = _fetch_int(path);
  free(path);
  if (rc == -1)
  {
    fprintf(stderr, "Could not fetch transaction number from ZooKeeper\n");
    return -1;
  }

  return rc;
}


static char * _lock_local_filename()
{
  static char lock_path[LONG_PATH_BUF_LEN];
  lock_path[LONG_PATH_BUF_LEN-1] = '\0';
  snprintf(lock_path, LONG_PATH_BUF_LEN-1, "%s/zk_lock", configuration.local_repo);
  return lock_path;
}

static int _save_lock(int lockvalue)
{
  int rc;
  FILE *fd;
  fd = fopen(_lock_local_filename(), "w");
  if (!fd) return -1;
  rc = fprintf(fd, "%d\n", lockvalue);
  if (rc < 0) return -1;
  fclose(fd);
  return 0;
}

static int _load_lock()
{
  int rc, lockvalue;
  FILE *fd;
  fd = fopen(_lock_local_filename(), "r");
  if (!fd) return -1;
  rc = fscanf(fd, "%d", &lockvalue);
  if (rc != 1)
  {
    if (!ferror(fd)) errno = EIO;
    fclose(fd);
    return -1;
  }
  fclose(fd);
  return lockvalue;
}

static int _delete_lock()
{
  return unlink(_lock_local_filename());
}

static int _convert_path_to_int(const char *path)
{
  int rc, the_int;
  char *dash = rindex(path, '-');
  if (!dash) return -1;
  rc = sscanf(dash+1, "%d", &the_int);
  if (rc != 1) return -1;
  return the_int;
}

//static int _lock_flag;
static pthread_mutex_t _acquire_lock_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t _acquire_lock_cond = PTHREAD_COND_INITIALIZER;

static void _acquire_lock_watcher(zhandle_t *zzh, int type, int state, const char *path, void *watcherCtx)
{
  //_lock_flag = 0;
  pthread_mutex_lock(&_acquire_lock_mutex); // grab the mutex
  pthread_cond_signal(&_acquire_lock_cond); // wake the thread that is waiting
  pthread_mutex_unlock(&_acquire_lock_mutex); // unlock
}

// Acquire a lock for modifying the repository
int acquire_lock(void)
{
  int rc;
  char *lock_path = _lock_path();
  char actual_path[PATH_BUF_LEN];
  actual_path[PATH_BUF_LEN-1] = '\0';
  char next_lowest_path[PATH_BUF_LEN];
  next_lowest_path[PATH_BUF_LEN-1] = '\0';
  struct String_vector children = { 0, 0 };
  int lockvalue, childvalue, lowest_childvalue, next_lowest_childvalue;

  // The algorithm here corresponds to the recipe at:
  // http://hadoop.apache.org/zookeeper/docs/current/recipes.html

  // 1. Create _lock_path() with sequence and ephemeral flags set
  rc = zoo_create(zh, lock_path, "", 0, &ZOO_OPEN_ACL_UNSAFE, ZOO_EPHEMERAL | ZOO_SEQUENCE,
                  actual_path, PATH_BUF_LEN-1);
  if (rc != ZOK)
  {
    fprintf(stderr, "Could not create ZooKeeper lock: %s\n", zerror(rc));
    return -1;
  }
  lockvalue = _convert_path_to_int(actual_path);
  if (lockvalue == -1)
  {
    fprintf(stderr, "Could not parse sequence number from the path returned by ZooKeeper\n");
    return -1;
  }

  while (1)
  {
    // 2. Call getChildren() on _lock_node() without setting the watch flag
    rc = zoo_get_children(zh, _lock_node(), 0, &children);
    if (rc != ZOK)
    {
      fprintf(stderr, "Could not get children of ZooKeeper lock node: %s\n", zerror(rc));
      return -1;
    }

    // 3. If pathname created in 1 has lowest sequence number suffix, we have the lock and may return
    lowest_childvalue = lockvalue;
    next_lowest_childvalue = -1;
    for (int i=0; i<children.count; i++)
    {
      childvalue = _convert_path_to_int(children.data[i]);
      if (childvalue == -1)
      {
        fprintf(stderr, "Could not parse sequence number from child node\n");
        return -1;
      }
      if (childvalue < lowest_childvalue) lowest_childvalue = childvalue;
      if (childvalue > next_lowest_childvalue && childvalue < lockvalue) next_lowest_childvalue = childvalue;
    }
    if (lowest_childvalue == lockvalue)
    {
      // Yay! We got the lock...
      rc = _save_lock(lockvalue);
      if (rc == -1)
      {
        perror("Could not save the lock number to the local repository");
        return -1;
      }
      return 0; // Success!
    }

    // Grab the mutex for signaling
    pthread_mutex_lock(&_acquire_lock_mutex);

    // 4. Otherwise call exists() with watch flag set on path with next lowest sequence number
    snprintf(next_lowest_path, PATH_BUF_LEN, "%s" LOCK_FORMAT_STR, _lock_path(), next_lowest_childvalue);
    //_lock_flag = 1;
    rc = zoo_wexists(zh, next_lowest_path, _acquire_lock_watcher, 0, 0); // second last arg is watcherCtx

    // 5. If exists() returns false, go to step 2. Otherwise wait for a notification
    // for the pathname before going to step 2.
    if (rc != ZOK)
    {
      if (rc == ZNONODE) continue; // seems we got the lock afterall
      fprintf(stderr, "Could not determine if ZooKeeper lock exists: %s\n", zerror(rc));
      pthread_mutex_unlock(&_acquire_lock_mutex); // be sure to unlock
      return -1;
    }

    // Wait for a notification
    pthread_cond_wait(&_acquire_lock_cond, &_acquire_lock_mutex);
    pthread_mutex_unlock(&_acquire_lock_mutex); // unlock
    //while (_lock_flag)
    //{
    //  sleep(1); // poll once per second
    //}

  }

  return -1; // should never get here!
}

// Release the lock for modifying the repository
int release_lock(void)
{
  int lockvalue;
  int rc;
  char lock_path[PATH_BUF_LEN];
  lock_path[PATH_BUF_LEN-1] = '\0';

  // Load the lock value from the local disk
  lockvalue = _load_lock();
  if (lockvalue == -1)
  {
    perror("Could not load lock file from local disk");
    return -1;
  }

  // Determine the lock path in ZooKeeper
  snprintf(lock_path, PATH_BUF_LEN, "%s" LOCK_FORMAT_STR, _lock_path(), lockvalue);

  // Remove the file from the local disk (This is done before releasing the lock
  // in ZooKeeper, in order to avoid a race condition where we might release the
  // lock and then delete the lock file belonging to another process that is
  // accessing this same repository.)
  _delete_lock();

  // Delete the lock path
  rc = zoo_delete(zh, lock_path, -1); // Should we check the node version here??
  if (rc != ZOK)
  {
    fprintf(stderr, "Could not delete lock in ZooKeeper: %s\n", zerror(rc));
    return -1;
  }

  return 0;
}

// Check if the lock is held (returns 1 if held, 0 if not held, -1 on error)
int check_lock(void)
{
  int lockvalue;
  int rc;
  char lock_path[PATH_BUF_LEN];
  lock_path[PATH_BUF_LEN-1] = '\0';

  // Load the lock value from the local disk
  lockvalue = _load_lock();
  if (lockvalue == -1)
  {
    // If we couldn't read the lock file, then assume that we don't have the lock
    return 0;
  }

  // Determine the lock path in ZooKeeper
  snprintf(lock_path, PATH_BUF_LEN, "%s" LOCK_FORMAT_STR, _lock_path(), lockvalue);

  // Check if the lock path exists
  rc = zoo_exists(zh, lock_path, 0, 0);
  if (rc == ZNONODE) return 0;
  else if (rc == ZOK) return 1;

  fprintf(stderr, "Could not determine if lock still exists: %s\n", zerror(rc));
  return -1;
}

// ZooKeeper watcher function
void watcher(zhandle_t *zzh, int type, int state, const char *path, void *watcherCtx)
{
  //printf("inside watcher type=%d state=%d path=%s\n", type, state, path);
  // FIXME: this function should not be empty!!
}

// Create a new ZooKeeper session
int create_session(void)
{
  zoo_set_debug_level(LOG_LEVEL_FOR_ZOO);
  zh = zookeeper_init(configuration.zk_addr, watcher, ZOOKEEPER_TIMEOUT, 0, 0, 0);
  if (!zh)
  {
    fprintf(stderr, "Could not create ZooKeeper session: %s\n", zerror(errno));
    return -1;
  }
  return 0;
}

// End the ZooKeeper session
int end_session(void)
{
  zookeeper_close(zh);
  return 0;
}

static char * _session_filename(void)
{
  static char session_file[PATH_BUF_LEN];
  session_file[PATH_BUF_LEN-1] = '\0';
  
  snprintf(session_file, PATH_BUF_LEN-1, "%s/zookeeper_session", configuration.local_repo);

  return session_file;
}

// Save the ZooKeeper client_id to the local repository
int save_session(void)
{
  int rc;
  FILE *fd;
  const clientid_t *clientid;

  // Get session 
  clientid = zoo_client_id(zh);
  if (!clientid)
  {
    fprintf(stderr, "Cannot get client_id from ZooKeeper\n");
    return -1;
  }

  // Write session info to disk
  fd = fopen(_session_filename(), "w");
  if (!fd)
  {
    perror("Cannot open session file for writing");
    return -1;
  }
  rc = fwrite(clientid, sizeof(clientid_t), 1, fd);
  if (rc != 1)
  {
    fprintf(stderr, "Cannot write session file\n");
    return -1;
  }
  fclose(fd);

  return 0;
}

// Delete the ZooKeeper client_id from the local repository
int delete_session(void)
{
  int rc;

  rc = unlink(_session_filename());
  if (rc == -1)
  {
    perror("Cannot delete session file");
    return -1;
  }

  return 0;
}

// Load an existing ZooKeeper client_id from the repository and connect the ZooKeeper session
int restore_session(void)
{
  int rc;
  clientid_t clientid;
  FILE *fd;

  // Get the session id from disk
  fd = fopen(_session_filename(), "r");
  if (!fd)
  {
    perror("Cannot open session file for reading");
    return -1;
  }
  rc = fread(&clientid, sizeof(clientid_t), 1, fd);
  if (rc != 1)
  {
    fprintf(stderr, "Cannot read session file\n");
    return -1;
  }
  fclose(fd);

  // Reconnect the ZooKeeper session
  zoo_set_debug_level(LOG_LEVEL_FOR_ZOO);
  zh = zookeeper_init(configuration.zk_addr, watcher, ZOOKEEPER_TIMEOUT, &clientid, 0, 0);
  if (!zh)
  {
    fprintf(stderr, "Could not restore ZooKeeper session: %s\n", zerror(errno));
    return -1;
  }

  return 0;
}

// Get an environment variable, printing a mesage and exiting if it is missing
char* get_env_or_die(char* env_name)
{
  char *value = getenv(env_name);
  if (!value)
  {
    fprintf(stderr, "Missing environment variable %s\n", env_name);
    exit(1);
  }
  return value;
}
