#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <string.h>

#include "s3ops.h"

#define CMD_PUT 1
#define CMD_GET 2
#define CMD_CLEAR 3

void print_usage()
{
  printf("\nUsage:\n");
  printf("  s3x put <bucket> <key> <local_filename>\n");
  printf("  s3x get <bucket> <key> <local_filename>\n");
  printf("  s3x clear <bucket>\n\n");
  exit(1);
}


// Get an environment variable, printing a mesage and exiting if it is missing
char* get_env_or_die(char* env_name)
{
  char *value = getenv(env_name);
  if (!value)
  {
    fprintf(stderr, "Missing environment variable %s\n", env_name);
    exit(1);
  }
  return value;
}


int main(int argc, char **argv)
{

  char *bucket = 0;
  char *key = 0;
  char *filename = 0;
  int cmd=0, rc, buf_len, fdno;
  void *buf;
  struct stat info;
  FILE *fd;

  if (argc < 2)
    print_usage();
  else if (!strcmp("clear", argv[1]))
    cmd = CMD_CLEAR;
  else if (!strcmp("put", argv[1]))
    cmd = CMD_PUT;
  else if (!strcmp("get", argv[1]))
    cmd = CMD_GET;
  else
    print_usage();

  if (cmd == CMD_CLEAR)
    if (argc == 3)
      bucket = argv[2];
    else
      print_usage();

  if (cmd == CMD_GET || cmd == CMD_PUT)
    if (argc == 5)
    {
      bucket = argv[2];
      key = argv[3];
      filename = argv[4];
    }
    else
      print_usage();

  // Initialize S3 credentials
  s3_init_credentials(get_env_or_die("AWS_KEY_ID"), get_env_or_die("AWS_SECRET_KEY"));

  // Perform the appropriate command
  if (cmd == CMD_CLEAR)
  {
    s3_clear_bucket(bucket);
  }
  else if (cmd == CMD_PUT)
  {
    
    // Open the file to upload
    fdno = open(filename, O_RDONLY);
    if (fdno == -1)
    {
      perror("Could not open file");
      exit(1);
    }
    rc = fstat(fdno, &info);
    if (rc == -1)
    {
      perror("Could not determine size of file");
      exit(1);
    }
    buf = mmap(0, info.st_size, PROT_READ, MAP_PRIVATE, fdno, 0);
    if (buf == MAP_FAILED)
    {
      perror("Could not map file into memory");
      exit(1);
    }

    // Now actually upload the file
    rc = s3_put_object(bucket, key, buf, info.st_size);
    if (rc == -1)
    {
      fprintf(stderr, "Could not put file into S3\n");
      munmap(buf, info.st_size);
      close(fdno);
      exit(1);
    }
    munmap(buf, info.st_size);
    close(fdno);

  }
  else if (cmd == CMD_GET)
  {

    // Fetch the file
    buf_len = s3_get_object(bucket, key, &buf);
    if (buf_len == -1)
    {
      fprintf(stderr, "Could not fetch file from S3\n");
      exit(1);
    }

    // Write the file to the local disk
    fd = fopen(filename, "w");
    if (!fd)
    {
      perror("Cannot open file for writing");
      free(buf);
      exit(1);
    }
    rc = fwrite(buf, buf_len, 1, fd);
    if (rc != 1)
    {
      perror("Cannot write file");
      free(buf);
      exit(1);
    }
    fclose(fd);
    free(buf);

  }

  return 0;
  
}
